// Truevision Targa file format specifications: http://www.dca.fee.unicamp.br/~martino/disciplinas/ea978/tgaffs.pdf

#include "../CFile.h"
#include <stdlib.h>
#include "TGA.h"

TGA tga;

void TGA::printError(const char* error, CFile::SFile* f)
{
    printf("TGA %s : %s\n", f->fileName, error);
    file.close(f);
    exit(EXIT_FAILURE);
}

Color* TGA::load(const char* filename, int* width, int* height)
{
    CFile::SFile*   f = file.open(filename, "rb");

    // read header
    SHeader header;
    readHeader(f, &header);

    // skip image ID
    if (header.IDLength != 0)
        file.seek(f, header.IDLength, SEEK_CUR);

    // read palette
    Color*  palette = readPalette(f, &header);

    // read image
    Color* image = readImage(f, &header, palette);
    *width  = header.imageWidth;
    *height = header.imageHeight;

    if (palette != NULL)
        delete[] palette;

    file.close(f);
    return image;
}

void TGA::readHeader(CFile::SFile* f, SHeader* h)
{
    h->IDLength = file.readU8(f);

    h->colorMapType = file.readU8(f);
    if (h->colorMapType > 1)
        printError("Wrong format", f);

    h->imageType = file.readU8(f);

    if (h->imageType != eIndexed &&
        h->imageType != eTrueColor &&
        h->imageType != eGray &&
        h->imageType != eIndexedRLE &&
        h->imageType != eTrueColorRLE &&
        h->imageType != eGrayRLE)
    {
        printError("Wrong format", f);
    }

    h->colorMap_firstEntryIndex = file.readU16(f);
    h->colorMap_length = file.readU16(f);
    h->colorMap_entrySize = file.readU8(f);

    if (h->colorMapType == 1 && h->colorMap_length != 0)
    {
        if (h->colorMap_entrySize != 15 &&
            h->colorMap_entrySize != 16 &&
            h->colorMap_entrySize != 24 &&
            h->colorMap_entrySize != 32)
        {
            printError("Wrong format", f);
        }
    }

    h->xOrigin = file.readU16(f);
    h->yOrigin = file.readU16(f);
    h->imageWidth  = file.readU16(f);
    h->imageHeight = file.readU16(f);
    h->pixelDepth = file.readU8(f);

    if (h->pixelDepth != 8 &&
        h->pixelDepth != 16 &&
        h->pixelDepth != 24 &&
        h->pixelDepth != 32)
    {
        printError("Wrong format", f);
    }

    Uint8 flags = file.readU8(f);

    if ((flags & 0xC0) != 0)
        printError("Wrong format", f);

    h->alphaBits = flags & 0xF;

    if (h->alphaBits != 0 &&
        h->alphaBits != 1 &&
        h->alphaBits != 8)
    {
        printError("Wrong format", f);
    }

    if ((flags & 0x10) == 0)
        h->hDirection = eLeftToRight;
    else
        h->hDirection = eRightToLeft;

    if ((flags & 0x20) == 0)
        h->vDirection = eBottomToTop;
    else
        h->vDirection = eTopToBottom;

    if (h->imageType == eIndexed && h->colorMap_length == 0)
        printError("Wrong format", f);

    if (h->imageType == eIndexedRLE && h->colorMap_length == 0)
        printError("Wrong format", f);
}

Color* TGA::readPalette(CFile::SFile* f, SHeader* h)
{
    Color*  palette = NULL;
    int bytesPerColor = (h->colorMap_entrySize + 7) / 8;

    if (h->imageType == eIndexed || h->imageType == eIndexedRLE)
    {
        palette = new Color[h->colorMap_length];

        for (int i = 0; i < h->colorMap_length; i++)
        {
            if (h->colorMap_entrySize == 15 || h->colorMap_entrySize == 16)
            {
                Uint16 value = file.readU16(f);
                palette[i].b = ( value        & 0x1F) << 3;
                palette[i].g = ((value >>  5) & 0x1F) << 3;
                palette[i].r = ((value >> 10) & 0x1F) << 3;
                palette[i].a = 255;
            }
            else if (h->colorMap_entrySize == 24)
            {
                palette[i].b = file.readU8(f);
                palette[i].g = file.readU8(f);
                palette[i].r = file.readU8(f);
                palette[i].a = 255;
            }
            else if (h->colorMap_entrySize == 32)
            {
                palette[i].b = file.readU8(f);
                palette[i].g = file.readU8(f);
                palette[i].r = file.readU8(f);
                palette[i].a = file.readU8(f);
            }
        }
    }
    else
    {
        file.seek(f, h->colorMap_length * bytesPerColor, SEEK_CUR);
    }

    return palette;
}

Color*  TGA::readImage(CFile::SFile* f, SHeader* h, Color* palette)
{
    int bytesPerPixel = (h->pixelDepth + 7) / 8;
    Color*  image = new Color[h->imageWidth * h->imageHeight];
    Uint8*  imageData = new Uint8[h->imageWidth * h->imageHeight * bytesPerPixel];
    int pixelIndex = 0;

    if (h->imageType == eIndexed ||
        h->imageType == eTrueColor ||
        h->imageType == eGray)
    {
        file.read(f, imageData, h->imageWidth * h->imageHeight * bytesPerPixel);
    }
    else
    {
        uncompress(f, h, imageData, bytesPerPixel);
    }

    for (int y = 0; y < h->imageHeight; y++)
        for (int x = 0; x < h->imageWidth; x++)
        {
            Color   c;

            // read the next pixel
            Uint32 pixel = 0;

            for (int i = 0; i < bytesPerPixel; i++)
                pixel |= imageData[pixelIndex++] << (i * 8);

            // decode color
            if (h->imageType == eIndexed || h->imageType == eIndexedRLE)
            {
                c = palette[pixel - h->colorMap_firstEntryIndex];
            }
            else if (h->imageType == eGray || h->imageType == eGrayRLE)
            {
                int g = pixel >> (h->pixelDepth - 8);
                c = Color(g, g, g, 255);
            }
            else
            {
                if (bytesPerPixel == 2)
                {
                    c.b = ( pixel        & 0x1F) << 3;
                    c.g = ((pixel >>  5) & 0x1F) << 3;
                    c.r = ((pixel >> 10) & 0x1F) << 3;
                    c.a = 255;
                }
                else if (bytesPerPixel == 3)
                {
                    c.b =  pixel        & 0xFF;
                    c.g = (pixel >>  8) & 0xFF;
                    c.r = (pixel >> 16) & 0xFF;
                    c.a = 255;
                }
                else if (bytesPerPixel == 4)
                {
                    c.b =  pixel        & 0xFF;
                    c.g = (pixel >>  8) & 0xFF;
                    c.r = (pixel >> 16) & 0xFF;
                    c.a = (pixel >> 24) & 0xFF;
                }
            }

            // handle image directions
            int destX, destY;

            if (h->hDirection == eLeftToRight)
                destX = x;
            else
                destX = h->imageWidth - 1 - x;

            if (h->vDirection == eTopToBottom)
                destY = y;
            else
                destY = h->imageHeight - 1 - y;

            image[destY * h->imageWidth + destX] = c;
        }

    return image;
}

void TGA::uncompress(CFile::SFile* f, SHeader* h, Uint8* imageData, int bytesPerPixel)
{
    int uncompSize = h->imageWidth * h->imageHeight * bytesPerPixel;
    int index = 0;

    while (index < uncompSize)
    {
        Uint8   counter = file.readU8(f);
        Uint8   counter2 = (counter & 0x7f) + 1;

        if (index + counter2 * bytesPerPixel > uncompSize)
            printError("Wrong format", f);

        if ((counter & 0x80) == 0)
        {
            // raw packet
            file.read(f, &imageData[index], counter2 * bytesPerPixel);
            index += counter2 * bytesPerPixel;
        }
        else
        {
            // run-length packet
            static Uint8 pixel[4];

            file.read(f, pixel, bytesPerPixel);

            for (int i = 0; i < counter2; i++)
                for (int j = 0; j < bytesPerPixel; j++)
                    imageData[index++] = pixel[j];
        }
    }
}

void TGA::save(const char* filename, int width, int height, Color* image, bool compress)
{
    CFile::SFile*   f = file.open(filename, "wb");

    writeHeader(f, width, height, compress);
    writeImage(f, width, height, image, compress);

    file.close(f);
}

void TGA::writeHeader(CFile::SFile* f, int width, int height, bool compress)
{
    file.writeU8(f, 0); // IDLength
    file.writeU8(f, 0); // colorMapType

    // imageType
    if (compress == false)
        file.writeU8(f, 2);
    else
        file.writeU8(f, 10);

    file.writeU16(f, 0); // colorMap_firstEntryIndex
    file.writeU16(f, 0); // colorMap_length
    file.writeU8(f, 0);  // colorMap_entrySize

    file.writeU16(f, 0);      // xOrigin
    file.writeU16(f, 0);      // yOrigin
    file.writeU16(f, width);  // imageWidth
    file.writeU16(f, height); // imageHeight
    file.writeU8(f, 32);      // pixelDepth

    file.writeU8(f, 8); // flags: 8 bits of alpha, left to right, bottom to top
}

void TGA::writeImage(CFile::SFile* f, int width, int height, Color* image, bool compress)
{
    if (compress == false)
    {
        for (int y = 0; y < height; y++)
            for (int x = 0; x < width; x++)
            {
                Color c = image[(height - 1 - y) * width + x];
                writePixel(f, c);
            }
    }
    else
    {
        for (int y = 0; y < height; y++)
        {
            Color* line = &image[(height - 1 - y) * width];
            compressLine(f, line, width);
        }
    }
}

void TGA::writePixel(CFile::SFile* f, Color c)
{
    file.writeU8(f, c.b);
    file.writeU8(f, c.g);
    file.writeU8(f, c.r);
    file.writeU8(f, c.a);
}

#define TGA_MIN_RUN     3
#define TGA_MAX_RUN     128
#define TGA_MIN_STRING  1
#define TGA_MAX_STRING  128

void TGA::compressLine(CFile::SFile* f, Color* input, int inputSize)
{
    int stringStart = 0;
    int runStart = 0;
    int currentPixel = 1;
    bool isInRun = false;

    while (true)
    {
        if (isInRun == false)
        {
            // string mode
            if (input[currentPixel] != input[runStart])
                runStart = currentPixel;

            int stringLength = runStart - stringStart;
            int runLength    = currentPixel + 1 - runStart;

            if (currentPixel == inputSize ||
                runLength    == TGA_MIN_RUN ||
                stringLength == TGA_MAX_STRING)
            {
                // emit string
                if (stringLength >= TGA_MIN_STRING)
                {
                    file.writeU8(f, stringLength - 1);

                    for (int i = 0; i < stringLength; i++)
                        writePixel(f, input[stringStart++]);
                }

                // go to run mode
                if (runLength == TGA_MIN_RUN)
                    isInRun = true;
            }
        }
        else
        {
            // run mode
            stringStart = currentPixel;

            int runLength = currentPixel - runStart;

            if (input[currentPixel] != input[runStart] ||
                currentPixel == inputSize ||
                runLength == TGA_MAX_RUN)
            {
                // emit run
                file.writeU8(f, 0x80 + runLength - 1);
                writePixel(f, input[runStart]);
                runStart = currentPixel;

                // go to string mode
                isInRun = false;
            }
        }

        if (currentPixel == inputSize)
            break;

        currentPixel++;
    }
}
